"""A class for the Lemke Howson algorithm"""
from nash.integer_pivoting import (make_tableau, non_basic_variables,
                                   pivot_tableau)

import numpy as np
from itertools import cycle


def shift_tableau(tableau, shape):
    """
    Shift a tableau to ensure labels of pairs of tableaux coincide

    Parameters
    ----------

        tableau: a numpy array
        shape: a tuple

    Returns
    -------

        tableau: a numpy array
    """
    return np.append(np.roll(tableau[:,:-1], shape[0], axis=1),
                     np.ones((shape[0], 1)), axis=1)

def tableau_to_strategy(tableau, basic_labels, strategy_labels):
    """
    Return a strategy vector from a tableau

    Parameters
    ----------

        tableau: a numpy array
        basic_labels: a set
        strategy_labels: a set

    Returns
    -------

        strategy: a numpy array
    """
    vertex = []
    for column in strategy_labels:
        if column in basic_labels:
            for i, row in enumerate(tableau[:, column]):
                if row != 0:
                    vertex.append(tableau[i, -1] / row)
        else:
            vertex.append(0)
    strategy = np.array(vertex)
    return strategy / sum(strategy)

def lemke_howson(A, B, initial_dropped_label=0):
    """
    Obtain the Nash equilibria using the Lemke Howson algorithm implemented
    using integer pivoting.

    Algorithm implemented here is Algorithm 3.6 of [Nisan2007]_.

    1. Start at the artificial equilibrium (which is fully labeled)
    2. Choose an initial label to drop and move in the polytope for which
       the vertex has that label to the edge
       that does not share that label. (This is implemented using integer
       pivoting)
    3. A label will now be duplicated in the other polytope, drop it in a
       similar way.
    4. Repeat steps 2 and 3 until have Nash Equilibrium.

    Parameters
    ----------

        initial_dropped_label: int

    Returns
    -------

        equilibria: A tuple.
    """

    if np.min(A) < 0:
        A = A + abs(np.min(A)) + 1
    if np.min(B) < 0:
        B = B + abs(np.min(B)) + 1

    # build tableax
    col_tableau = make_tableau(A)
    col_tableau = shift_tableau(col_tableau, A.shape)
    row_tableau = make_tableau(B.transpose())
    full_labels = set(range(sum(A.shape)))

    if initial_dropped_label in non_basic_variables(row_tableau):
        tableux = cycle((row_tableau, col_tableau))
    else:
        tableux = cycle((col_tableau, row_tableau))

    # First pivot (to drop a label)
    entering_label = pivot_tableau(next(tableux), initial_dropped_label)
    while non_basic_variables(row_tableau).union(non_basic_variables(col_tableau)) != full_labels:
        entering_label = pivot_tableau(next(tableux), next(iter(entering_label)))

    row_strategy = tableau_to_strategy(row_tableau, non_basic_variables(col_tableau),
                                       range(A.shape[0]))
    col_strategy = tableau_to_strategy(col_tableau, non_basic_variables(row_tableau),
                                       range(A.shape[0], sum(A.shape)))

    return row_strategy, col_strategy



"""A class for a normal form game"""
import numpy as np
from .algorithms.lemke_howson import lemke_howson
from .algorithms.support_enumeration import support_enumeration
from .algorithms.vertex_enumeration import vertex_enumeration
from itertools import combinations


class Game:
    """
    A class for a normal form game.

    Parameters
    ----------

        - A, B: 2 dimensional list/arrays representing the payoff matrices for
          non zero sum games.
        - A: 2 dimensional list/array representing the payoff matrix for a
          zero sum game.
    """
    def __init__(self, *args):
        if len(args) == 2:
            self.payoff_matrices = tuple([np.asarray(m) for m in args])
        if len(args) == 1:
            self.payoff_matrices = np.asarray(args[0]), -np.asarray(args[0])
        self.zero_sum = np.array_equal(self.payoff_matrices[0],
                                       -self.payoff_matrices[1])

    def __repr__(self):
        if self.zero_sum:
            tpe = "Zero sum"
        else:
            tpe = "Bi matrix"
        return """{} game with payoff matrices:

Row player:
{}

Column player:
{}""".format(tpe, *self.payoff_matrices)

    def __getitem__(self, key):
        row_strategy, column_strategy = key
        return np.array([np.dot(row_strategy, np.dot(m, column_strategy))
                         for m in self.payoff_matrices])

    def vertex_enumeration(self):
        """
        Obtain the Nash equilibria using enumeration of the vertices of the best
        response polytopes.

        Algorithm implemented here is Algorithm 3.5 of [Nisan2007]_.

        1. Build best responses polytopes of both players
        2. For each vertex pair of both polytopes
        3. Check if pair is fully labelled
        4. Return the normalised pair

        Returns
        -------

            equilibria: A generator.
        """
        return vertex_enumeration(*self.payoff_matrices)

    def support_enumeration(self):
        """
        Obtain the Nash equilibria using support enumeration.

        Algorithm implemented here is Algorithm 3.4 of [Nisan2007]_.

        1. For each k in 1...min(size of strategy sets)
        2. For each I,J supports of size k
        3. Solve indifference conditions
        4. Check that have Nash Equilibrium.

        Returns
        -------

            equilibria: A generator.
        """
        return support_enumeration(*self.payoff_matrices)

    def lemke_howson(self, initial_dropped_label):
        """
        Obtain the Nash equilibria using the Lemke Howson algorithm implemented
        using integer pivoting.

        Algorithm implemented here is Algorithm 3.6 of [Nisan2007]_.

        1. Start at the artificial equilibrium (which is fully labeled)
        2. Choose an initial label to drop and move in the polytope for which
           the vertex has that label to the edge
           that does not share that label. (This is implemented using integer
           pivoting)
        3. A label will now be duplicated in the other polytope, drop it in a
           similar way.
        4. Repeat steps 2 and 3 until have Nash Equilibrium.

        Parameters
        ----------

            initial_dropped_label: int

        Returns
        -------

            equilibria: A tuple.
        """
        return lemke_howson(*self.payoff_matrices,
                            initial_dropped_label=initial_dropped_label)



"""A class for a normal form game"""
import numpy as np
from itertools import chain, combinations


def powerset(n):
    """
    A power set of range(n)

    Based on recipe from python itertools documentation:

    https://docs.python.org/2/library/itertools.html#recipes
    """
    return chain.from_iterable(combinations(range(n), r) for r in range(n + 1))


def solve_indifference(A, rows=None, columns=None):
    """
    Solve the indifference for a payoff matrix assuming support for the
    strategies given by columns

    Finds vector of probabilities that makes player indifferent between
    rows.  (So finds probability vector for corresponding column player)

    Parameters
    ----------

        A: a 2 dimensional numpy array (A payoff matrix for the row player)
        rows: the support played by the row player
        columns: the support player by the column player

    Returns
    -------

        A numpy array:
        A probability vector for the column player that makes the row
        player indifferent. Will return False if all entries are not >= 0.
    """
    # Ensure differences between pairs of pure strategies are the same
    M = (A[np.array(rows)] - np.roll(A[np.array(rows)], 1, axis=0))[:-1]

    # Columns that must be played with prob 0
    zero_columns = set(range(A.shape[1])) - set(columns)

    if zero_columns != set():
        M = np.append(M, [[int(i == j) for i, col in enumerate(M.T)]
                          for j in zero_columns], axis=0)

    # Ensure have probability vector
    M = np.append(M, np.ones((1, M.shape[1])), axis=0)
    b = np.append(np.zeros(len(M) - 1), [1])

    try:
        prob = np.linalg.solve(M, b)
        if all(prob >= 0):
            return prob
        return False
    except np.linalg.linalg.LinAlgError:
        return False

def potential_support_pairs(A, B):
    """
    A generator for the potential support pairs

    Returns
    -------

        A generator of all potential support pairs
    """
    p1_num_strategies, p2_num_strategies = A.shape
    for support1 in (s for s in powerset(p1_num_strategies) if len(s) > 0):
        for support2 in (s for s in powerset(p2_num_strategies)
                         if len(s) == len(support1)):
            yield support1, support2

def indifference_strategies(A, B):
    """
    A generator for the strategies corresponding to the potential supports

    Returns
    -------

        A generator of all potential strategies that are indifferent on each
        potential support. Return False if they are not valid (not a
        probability vector OR not fully on the given support).
    """
    for pair in potential_support_pairs(A, B):
        s1 = solve_indifference(B.T, *(pair[::-1]))
        s2 = solve_indifference(A, *pair)

        if obey_support(s1, pair[0]) and obey_support(s2, pair[1]):
            yield s1, s2, pair[0], pair[1]

def obey_support(strategy, support):
    """
    Test if a strategy obeys its support

    Parameters
    ----------

        strategy: a numpy array
            A given strategy vector
        support: a numpy array
            A strategy support

    Returns
    -------

        A boolean: whether or not that strategy does indeed have the given
        support
    """
    if strategy is False:
        return False
    if not all((i in support and value > 0) or
               (i not in support and value <= 0)
               for i, value in enumerate(strategy)):
        return False
    return True

def is_ne(strategy_pair, support_pair, payoff_matrices):
    """
    Test if a given strategy pair is a pair of best responses

    Parameters
    ----------

        strategy_pair: a 2-tuple of numpy arrays
        support_pair: a 2-tuple of numpy arrays
    """
    A, B = payoff_matrices
    # Payoff against opponents strategies:
    u = strategy_pair[1].reshape(strategy_pair[1].size, 1)
    row_payoffs = np.dot(A, u)

    v = strategy_pair[0].reshape(strategy_pair[0].size, 1)
    column_payoffs = np.dot(B.T, v)

    # Pure payoffs on current support:
    row_support_payoffs = row_payoffs[np.array(support_pair[0])]
    column_support_payoffs = column_payoffs[np.array(support_pair[1])]

    return (row_payoffs.max() == row_support_payoffs.max() and
            column_payoffs.max() == column_support_payoffs.max())

def support_enumeration(A, B):
    """
    Obtain the Nash equilibria using support enumeration.

    Algorithm implemented here is Algorithm 3.4 of [Nisan2007]_

    1. For each k in 1...min(size of strategy sets)
    2. For each I,J supports of size k
    3. Solve indifference conditions
    4. Check that have Nash Equilibrium.

    Returns
    -------

        equilibria: A generator.
    """
    return ((s1, s2)
            for s1, s2, sup1, sup2 in indifference_strategies(A, B)
            if is_ne((s1, s2), (sup1, sup2), (A, B)))



"""A class for the vertex enumeration algorithm"""
from nash.polytope import build_halfspaces, non_trivial_vertices

import numpy as np
from itertools import product

def vertex_enumeration(A, B):
    """
    Obtain the Nash equilibria using enumeration of the vertices of the best
    response polytopes.

    Algorithm implemented here is Algorithm 3.5 of [Nisan2007]_

    1. Build best responses polytopes of both players
    2. For each vertex pair of both polytopes
    3. Check if pair is fully labelled
    4. Return the normalised pair

    Returns
    -------

        equilibria: A generator.
    """

    if np.min(A) < 0:
        A = A + abs(np.min(A))
    if np.min(B) < 0:
        B = B + abs(np.min(B))

    number_of_row_strategies, row_dimension = A.shape
    max_label = number_of_row_strategies + row_dimension
    full_labels = set(range(max_label))

    row_halfspaces = build_halfspaces(B.transpose())
    col_halfspaces = build_halfspaces(A)

    for row_v, row_l in non_trivial_vertices(row_halfspaces):
        adjusted_row_l = set((label + number_of_row_strategies) % (max_label)
                             for label in row_l)

        for col_v, col_l in non_trivial_vertices(col_halfspaces):
            if adjusted_row_l.union(col_l) == full_labels:
                yield row_v / sum(row_v), col_v / sum(col_v)



"""
A class for integer pivoting. Used for an implementation of the Lemke Howson
algorithm.
"""
import numpy as np

def make_tableau(M):
    """
    Make a tableau for the given matrix M.

    This tableau corresponds to the polytope of the form:

       Mx <= 1 and x >= 0
    """
    return np.append(np.append(M, np.eye(M.shape[0]), axis=1),
                     np.ones((M.shape[0], 1)), axis=1)


def find_pivot_row(tableau, column):
    """
    Find the index of the row to pivot.

    Identifies the row to pivot by performing a minimum ratio test. (In fact
    implemented to calculate the maximum ratio test to avoid divide by zero
    errors).
    """
    return np.argmax(tableau[:, column] / tableau[:, -1])


def non_basic_variables(tableau):
    """
    Identifies the non basic variables of a tableau,
    these correspond to the labels.
    """
    columns = tableau[:,:-1].transpose()
    return set(np.where([np.count_nonzero(col) != 1 for col in columns])[0])


def pivot_tableau(tableau, column_index):
    """
    Pivots the tableau and returns the dropped label
    """
    original_labels = non_basic_variables(tableau)
    pivot_row_index = find_pivot_row(tableau, column_index)
    pivot_element = tableau[pivot_row_index, column_index]

    for i, _ in enumerate(tableau):
        if i != pivot_row_index:
            tableau[i, :] = (
                tableau[i, :] * pivot_element -
                tableau[pivot_row_index, :] * tableau[i, column_index])

    return non_basic_variables(tableau) - original_labels



from setuptools import setup, find_packages
import unittest
import doctest
import os

# Read in the version number
exec(open('src/nash/version.py', 'r').read())

requirements = ["numpy>=1.12.1", "scipy>=0.19.0"]

def test_suite():
    """Discover all tests in the tests dir"""
    test_loader = unittest.TestLoader()
    # Read in unit tests
    test_suite = test_loader.discover('tests')

    # Doctest all md and rst files
    for root, dirs, files in os.walk("."):
        for f in files:
            if f.endswith(".rst") or f.endswith(".md"):
                test_suite.addTests(
                     doctest.DocFileSuite(os.path.join(root, f),
                                          optionflags=doctest.ELLIPSIS))

    return test_suite

setup(
    name='nashpy',
    version=__version__,
    install_requires=requirements,
    author='Vince Knight, James Campbell',
    author_email=('knightva@cardiff.ac.uk'),
    packages=find_packages('src'),
    package_dir={"": "src"},
    test_suite='setup.test_suite',
    url='',
    license='The MIT License (MIT)',
    description='A library to compute equilibria of 2 player normal form games',
)
